﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace XNA4
{
    public class ModelNode : DrawableGameComponent
    {
        Game game;

        /// <summary>
        /// model object/name
        /// </summary>
        Model mesh;
        string name;

        /// <summary>
        /// model position/scale/rotation
        /// </summary>
        public Vector3 Position;
        public float Scale;
        public Quaternion Rotation;

        /// <summary>
        /// model matrix
        /// </summary>
        public Matrix World;
        public Matrix View;
        public Matrix Projection;
        public bool Toon = false;
        Effect ToonEffect;
        Texture2D ToonTexture;

        // bounding volume
        //BoundingBox volume;
        public BoundingSphere Sphere;
        //GraphicsDevice gd;

        /// <summary>
        /// ctor
        /// </summary>
        public ModelNode(Game game, string modelAssetName)
            : this(game, modelAssetName, Vector3.Zero, 1.0f, Quaternion.Identity, false)
        {
        }

        public ModelNode(Game game, string modelAssetName, Vector3 position, float scale, Quaternion rotate, bool toon)
            : base(game)
        {
            this.game = game;

            // model load
            this.name = modelAssetName;
            if (game.Content != null && this.mesh == null && this.name != string.Empty)
                this.mesh = game.Content.Load<Model>(this.name);

            // effect
            if (game.Content != null && toon)
            {
                this.ToonEffect = game.Content.Load<Effect>("Effect\\Toon");
                this.ToonTexture = game.Content.Load<Texture2D>("Effect\\HelicopterTexture");
            }

            // model transformation
            this.Position = position;
            this.Scale = scale;
            this.Rotation = rotate;
            
            // bounding volume     
            //this.volume = UpdateBoundingBox();
            this.Sphere = UpdateBoundingSphere();
            System.Diagnostics.Trace.WriteLine("boundingsphere = " + this.Sphere);
            BoundingSphereRenderer.InitializeGraphics(game.GraphicsDevice, 30);
        }

        /// <summary>
        /// Allows the game component to perform any initialization it needs to before starting
        /// to run.  This is where it can query for any required services and load content.
        /// </summary>
        public override void Initialize()
        {
            base.Initialize();
        }

        /// <summary>
        /// Allows the game component to update itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        public override void Update(GameTime gameTime)
        {
        }

        // model's bounding box
        protected BoundingBox UpdateBoundingBox() 
        {     
            // initialize minimum and maximum corners of the bounding box to max and min values     
            Vector3 min = new Vector3(float.MaxValue, float.MaxValue, float.MaxValue);     
            Vector3 max = new Vector3(float.MinValue, float.MinValue, float.MinValue);      
            // For each mesh of the model     
            foreach (ModelMesh mesh in this.mesh.Meshes)     
            {         
                foreach (ModelMeshPart meshPart in mesh.MeshParts)         
                {             
                    // Vertex buffer parameters             
                    int vertexStride = meshPart.VertexBuffer.VertexDeclaration.VertexStride;             
                    int vertexBufferSize = meshPart.NumVertices * vertexStride;              
                    // Get vertex data as float             
                    float[] vertexData = new float[vertexBufferSize / sizeof(float)];             
                    meshPart.VertexBuffer.GetData<float>(vertexData);              
                    // Iterate through vertices (possibly) growing bounding box, all calculations are done in world space             
                    for (int i = 0; i < vertexBufferSize / sizeof(float); i += vertexStride / sizeof(float))             
                    {                 
                        Vector3 transformedPosition = Vector3.Transform(new Vector3(vertexData[i], vertexData[i + 1], vertexData[i + 2]), World);                  
                        min = Vector3.Min(min, transformedPosition);                 
                        max = Vector3.Max(max, transformedPosition);             
                    }         
                }     
            }      
            // Create and return bounding box     
            return new BoundingBox(min, max); 
        }

        // model's bounding sphere
        protected BoundingSphere UpdateBoundingSphere()
        {
            BoundingSphere sphere = new BoundingSphere();

            foreach (ModelMesh mesh in this.mesh.Meshes)
            {
                if (sphere.Radius == 0)
                    sphere = mesh.BoundingSphere;
                else
                    sphere = BoundingSphere.CreateMerged(sphere, mesh.BoundingSphere); 
            }

            sphere.Center = this.Position;
            sphere.Radius *= this.Scale;
            return sphere;
        }

        public override void Draw(GameTime gameTime)
        {
            if (this.mesh == null) return;

            Matrix[] transforms = new Matrix[this.mesh.Bones.Count];
            this.mesh.CopyAbsoluteBoneTransformsTo(transforms);

            if (this.ToonEffect != null)
            {
                foreach (ModelMesh mesh in this.mesh.Meshes)
                {
                    foreach (ModelMeshPart part in mesh.MeshParts)
                    {
                        part.Effect = ToonEffect;
                        ToonEffect.Parameters["World"].SetValue(mesh.ParentBone.Transform * Matrix.CreateScale(this.Scale) * Matrix.CreateFromQuaternion(this.Rotation) * Matrix.CreateTranslation(this.Position) * this.World);
                        ToonEffect.Parameters["View"].SetValue(this.View);
                        ToonEffect.Parameters["Projection"].SetValue(this.Projection);
                        ToonEffect.Parameters["WorldInverseTranspose"].SetValue(
                                                Matrix.Transpose(Matrix.Invert(mesh.ParentBone.Transform * Matrix.CreateScale(this.Scale) * Matrix.CreateFromQuaternion(this.Rotation) * Matrix.CreateTranslation(this.Position) * this.World)));
                        ToonEffect.Parameters["Texture"].SetValue(ToonTexture);
                    }
                    // draw the mesh, will use the effects set above.
                    mesh.Draw();
                }
            }
            else
            {
                foreach (ModelMesh mesh in this.mesh.Meshes)
                {
                    // this is where the mesh orientation is set, as well as our camera and projection
                    foreach (BasicEffect effect in mesh.Effects)
                    {
                        effect.EnableDefaultLighting();
                        effect.World = transforms[mesh.ParentBone.Index] * Matrix.CreateScale(this.Scale) * Matrix.CreateFromQuaternion(this.Rotation) * Matrix.CreateTranslation(this.Position) * this.World;
                        effect.View = this.View;
                        effect.Projection = this.Projection;
                    }
                    // draw the mesh, will use the effects set above.
                    mesh.Draw();
                }
            }
            // draw bounding volume
            DrawBoundingVolume();
        }

        // draw bounding volume
        public virtual void DrawBoundingVolume()
        {
            BoundingSphereRenderer.Render(this.Sphere, this.game.GraphicsDevice, this.World, this.View, this.Projection, Color.White);
        }

    }
}